@@ -1,7 +1,7 @@
 
-__version__ = "5.5"
+__version__ = "5.7"
 docs = r"https://crocodile.readthedocs.io/en/latest/"
 git = r"https://github.com/thisismygitrepo/crocodile"
 release_notes = """
-added inplace method to Path
+new methods to P
 """
