@@ -1,6 +1,15 @@
 """
 A collection of classes extending the functionality of Python's builtins.
 email programmer@usa.com
+
+Crocodile Philosophy:
+Make Python even friendlier, by making available the common functionality for everyday use. Specifically,
+path management, file management, and other common tasks.
+A the risk of vandalizing the concept, to make Python more MATLAB-like, in that more libraries are loaded up at
+start time than basic arithmetic, but just enought to make it more useful for everyday errands.
+Thus, the terseness of Crocodile makes Python REPL a proper shell
+
+The focus is on ease of use, not efficiency.
 """
 
 # Typing
@@ -66,8 +75,13 @@ def randstr(length=10, lower=True, upper=True, digits=True, punctuation=False, s
     return result_str
 
 
-def assert_package_installed(package):
-    """imports a package and installs it if not."""
+def validate_name(astring, replace='_'):
+    import re
+    return re.sub(r'^(?=\d)|\W', replace, str(astring))
+
+
+def install_n_import(package):
+    """imports a package and installs it first if not."""
     try:
         pkg = __import__(package)
         return pkg
@@ -125,14 +139,14 @@ def save_decorator(ext=""):
     """Apply default paths, add extension to path, print the saved file path"""
 
     def decorator(func):
-        def wrapper(obj, path=None, verbose=True, **kwargs):
+        def wrapper(obj, path=None, verbose=True, add_suffix=True, **kwargs):
             if path is None:
                 path = Path.home().joinpath("tmp_results").joinpath(randstr() + ext)
                 print(f"tb.core: Warning: Path not passed to {func}. "
                       f"A default path has been chosen: {path.absolute().as_uri()}")
                 # raise ValueError
             else:
-                if not str(path).endswith(ext):
+                if add_suffix and not str(path).endswith(ext):
                     path = Path(str(path) + ext)
                     print(f"tb.core: Warning: suffix {ext} is added to path passed {path.as_uri()}")
                 else:
@@ -145,9 +159,7 @@ def save_decorator(ext=""):
                 rep = rep if len(rep) < 50 else rep[:10] + "... "
                 print(f"SAVED {rep}  @ `{path.absolute().as_uri()}` |  Directory: `{path.parent.absolute().as_uri()}`")
             return path
-
         return wrapper
-
     return decorator
 
 
@@ -220,7 +232,7 @@ class Save:
             dill.dump(obj, file, recurse=r, **kwargs)
 
     @staticmethod
-    def pickle_s(obj):
+    def pickles(obj):
         binary = dill.dumps(obj)
         return binary
 
@@ -253,31 +265,45 @@ class Base(object):
         Path(path).write_text(file.read_text())
         return Path(path) if type(path) is str else path  # path could be tb.P, better than Path
 
-    def save(self, path=None, itself=True, r=False, include_code=False):
+    def save(self, path=None, itself=True, r=False, include_code=False, add_suffix=True):
         """Pickles the object.
         :param path: destination file.
-        :param itself: `itself` means the object (self) will be pickled straight away. This is the default behaviour,
-        however, it requires (in case of composed objects) that every sub-object is well-behaved and has the appropriate
-        state methods implemented. The alternative to this option (itself=False) is to save __dict__ only
-         (assuming it is pure data rather than code, otherwise recusive flag must be set), then the class itself is
-          required later and the `from_pickled_state` method should be used to reload the instance again.
-          The disadvantage of this method is that __init__ method will be used again at reconstruction time
-           of the object before the attributes are monkey-patched.
-           It is very arduous to design __init__ method that is convenient (uses plethora of
-          default arguments) and works at the same time with no input at reconstruction time.
-          # Use only for classes with whacky behaviours or too expensive to redesign
-            # methodology: 1- Save state, 2- save code. 3- initialize from __init__, 4- populate __dict__
+        :param itself:
+        * `itself` is True: the object (self) will be pickled altogether.
+            * Ups:
+                * convenient.
+                * Dill can, by itself import all the required classes to __main__ IF the object is restored
+                 while directory is @ the same location object was created, thus,
+            * Downs:
+                * It requires (in case of composed objects) that every sub-object is well-behaved and has the appropriate
+                  state methods implemented.
+                * The libraries imported must be present at load time.
+        * `itself` is False: means to save __getstate__.
+            * Ups:
+                * very safe and unlikely to cause import errors at load time.
+                * User is responsible for loading up the class.
+            * Downs: the class itself is required later and the `from_pickled_state` method should be used to reload the instance again.
+            * __init__ method will be used again at reconstruction time of the object before the attributes are monkey-patched.
+            * It is very arduous to design __init__ method that is convenient (uses plethora of
+              default arguments) and works at the same time with no input at reconstruction time.
+
         :param include_code: `save_code` will be called.
         :param r: recursive flag.
+            * If attributes are not data, but rather objects, then, this flag should be set to True.
+            * The recusive flag is particularly relevant when `itself` is False and __dict__ is composed of objects.
+            * In pickling the object (itself=True), the recursive flag is the default.
+        :param add_suffix: if True, the suffixes `.pkl` and `.py` will be added to the file name.
 
-        * Dill package manages to resconstruct the object by loading up all the appropriate libraries again
-        IF the object is restored while directory is @ the same location object was created, thus,
-        no need for saving code or even reloading it.
-        * Beware of the security risk involved in pickling objects that reference sensitive information like tokens and
-        passwords. The best practice is to pass them again at load time.
+        * Tip: whether pickling the class or its data alone, always implement __getstate__ appropriately to
+        avoid security risk involved in pickling objects that reference sensitive information like tokens and
+        passwords.
         """
-        path = str(path or Path.home().joinpath(f"tmp_results/tmpfiles/{randstr()}"))
-        path = Path(path + "." + self.__class__.__name__ + ("" if itself else ".dat"))
+        path = path or Path.home().joinpath(f"tmp_results/tmpfiles/{randstr()}")
+        if add_suffix:
+            path = str(path)
+            path += "" if self.__class__.__name__ in path else ("." + self.__class__.__name__)
+            path += "" if (itself or ".dat" in path) else ".dat"
+        path = Path(path)
         # Fruthermore, .zip or .pkl will be added later depending on `include_code` value, warning will be raised.
 
         # Choosing what object to pickle:
@@ -297,22 +323,25 @@ class Base(object):
             temp_path = Path().home().joinpath(f"tmp_results/zipping/{randstr()}")
             temp_path.mkdir(parents=True, exist_ok=True)
             self.save_code(path=temp_path.joinpath(f"source_code_{randstr()}.py"))
-            Save.pickle(path=temp_path.joinpath("class_data"), obj=obj, r=r, verbose=False)
+            Save.pickle(path=temp_path.joinpath("class_data"), obj=obj, r=r, verbose=False, add_suffix=add_suffix)
             import shutil
             result_path = shutil.make_archive(base_name=str(path), format="zip",
                                               root_dir=str(temp_path), base_dir=".")
             result_path = Path(result_path)
             print(f"Code and data for the object ({repr(obj)}) saved @ "
-                  f"{result_path.as_uri()}, Directory: {result_path.parent.as_uri()}")
+                  f"`{result_path.as_uri()}`, Directory: `{result_path.parent.as_uri()}`")
         else:
-            result_path = Save.pickle(obj=obj, path=path, r=r, verbose=False)
-            print(f"{'Data of' if itself else ''} Object ({repr(obj)}) saved @ "
-                  f"{result_path.absolute().as_uri()}, Directory: {result_path.parent.absolute().as_uri()}")
+            result_path = Save.pickle(obj=obj, path=path, r=r, verbose=False, add_suffix=add_suffix)
+            rep = repr(obj)
+            rep = rep if len(rep) < 50 else rep[:10] + "... "
+            print(f"{'Data of' if itself else ''} Object ({rep}) saved @ "
+                  f"`{result_path.absolute().as_uri()}`, Directory: `{result_path.parent.absolute().as_uri()}`")
         return result_path
 
     @classmethod
     def from_saved(cls, path, *args, r=False, scope=None, **kwargs):
         """Works in conjuction with save_pickle when `itself`=False, i.e. only state of object is pickled.
+        # methodology: 1- Save state, 2- save code. 3- initialize from __init__, 4- populate __dict__
 
         :param path: points to where the `data` of this class is saved
         :param r: recursive flag. If set to True, then, the directory containing the file `path` will be searched for
@@ -329,7 +358,7 @@ class Base(object):
          e.g. do NOT create some expensive attribute is this flag is raised because it will be obtained later.
         """
 
-        # ============================= step 1: load up the data
+        # ============================= step 1: unpickle the data
         assert ".dat." in str(path), f"Are you sure the path {path} is pointing to pickeld state of {cls}?"
         data = dill.loads(Path(path).read_bytes())
         # ============================= step 2: initialize the class
@@ -449,8 +478,8 @@ class Base(object):
     def print(self, typeinfo=False):
         Struct(self.__dict__).print(dtype=typeinfo)
 
-    def viz_heirarchy(self, depth=3, obj=None, filt=None):
-        import objgraph
+    def viz_composition_heirarchy(self, depth=3, obj=None, filt=None):
+        objgraph = install_n_import("objgraph")
         import tempfile
         filename = Path(tempfile.gettempdir()).joinpath("graph_viz_" + randstr() + ".png")
         objgraph.show_refs([self] if obj is None else [obj], max_depth=depth, filename=str(filename), filter=filt)
@@ -460,8 +489,9 @@ class Base(object):
         return filename
 
 
-class List(list, Base):
+class List(Base, list):
     """Use this class to keep items of the same type.
+    # Inheriting from Base gives save method.
     """
 
     # =============================== Constructor Methods ====================
@@ -469,12 +499,11 @@ class List(list, Base):
         super().__init__()
         self.list = list(obj_list) if obj_list is not None else []
 
-    def insert(self, __index: int, __object):
-        self.list.insert(__index, __object)
-        return self
-
     def __bool__(self):
         return bool(self.list)
+    
+    def __contains__(self, key):
+        return key in self.list
 
     @classmethod
     def from_copies(cls, obj, count):
@@ -580,20 +609,6 @@ class List(list, Base):
         """Select at random"""
         return self[np.random.choice(len(self), size, replace=replace, p=p)]
 
-    def to_struct(self, key_val=None):
-        """
-        :param key_val: function that returns (key, value) pair.
-        :return:
-        """
-        if key_val is None:
-            def key_val(x):
-                return str(x), x
-        else:
-            key_val = self.evalstr(key_val)
-        # return Struct.from_keys_values_pairs(self.apply(key_val))
-        # removed for disentanglement
-        return dict(self.apply(key_val))
-
     # def find(self, patt, match="fnmatch"):
     #     """Looks up the string representation of all items in the list and finds the one that partially matches
     #     the argument passed. This method is a short for ``self.filter(lambda x: string_ in str(x))`` If you need more
@@ -694,7 +709,7 @@ class List(list, Base):
             self.apply(lambda x: x.apply(func, *args, other=other, jobs=jobs, depth=depth, **kwargs))
 
         func = self.evalstr(func, expected='func')
-        tqdm = assert_package_installed("tqdm").tqdm
+        tqdm = install_n_import("tqdm").tqdm
         if other is None:
             iterator = self.list if not verbose else tqdm(self.list, desc=desc)
             if jobs:
@@ -742,6 +757,10 @@ class List(list, Base):
                 result.append(item)
         return result
 
+    def insert(self, __index: int, __object):
+        self.list.insert(__index, __object)
+        return self
+
     def print(self, nl=1, sep=False, style=repr):
         for idx, item in enumerate(self.list):
             print(f"{idx:2}- {style(item)}", end=' ')
@@ -756,17 +775,16 @@ class List(list, Base):
 
     def to_dataframe(self, names=None, minimal=False, obj_included=True):
         """
-
         :param names: path of each object.
         :param minimal: Return Dataframe structure without contents.
         :param obj_included: Include a colum for objects themselves.
         :return:
         """
-        # DisplayData.set_pandas_display()  # removed for disentanglement
+        import pandas as pd
+        # Display.set_pandas_display()
         columns = list(self.list[0].__dict__.keys())
         if obj_included or names:
             columns = ['object'] + columns
-        import pandas as pd
         df = pd.DataFrame(columns=columns)
         if minimal:
             return df
@@ -783,6 +801,10 @@ class List(list, Base):
                 df.loc[i] = list(self.list[i].__dict__.values())
         return df
 
+    def to_list(self):
+        """Matches `to_list` from Pandas Series."""
+        return self.list
+
     def to_numpy(self):
         return self.np
 
@@ -790,22 +812,27 @@ class List(list, Base):
     def np(self):
         return np.array(self.list)
 
+    def to_struct(self, key_val=None):
+        """
+        :param key_val: function that returns (key, value) pair.
+        :return:
+        """
+        if key_val is None:
+            def key_val(x):
+                return str(x), x
+        else:
+            key_val = self.evalstr(key_val)
+        return Struct.from_keys_values_pairs(self.apply(key_val))
+        # return dict(self.apply(key_val))
 
-class Struct(dict):  # inheriting from dict gives `get` method.
+
+class Struct(Base, dict):
     """Use this class to keep bits and sundry items.
     Combines the power of dot notation in classes with strings in dictionaries to provide Pandas-like experience
+    # inheriting from dict gives `get` method, should give `__contains__` but not working.
+    # Inheriting from Base gives `save` method.
     """
 
-    def save_json(self, path=None):
-        path = Save.json(obj=self.__dict__, path=path)
-        return path
-
-    def save_yaml(self, path=None):
-        return Save.yaml(obj=self.__dict__, path=path)
-
-    def __len__(self):
-        return len(self.keys())
-
     def __init__(self, dictionary=None, **kwargs):
         """
         :param dictionary: a dict, a Struct, None or an object with __dict__ attribute.
@@ -827,15 +854,12 @@ class Struct(dict):  # inheriting from dict gives `get` method.
             final_dict.update(kwargs)
         self.__dict__ = final_dict
 
-    def to_default(self, default=lambda: None):
-        from collections import defaultdict
-        tmp2 = defaultdict(default)
-        tmp2.update(self.__dict__)
-        self.__dict__ = tmp2
-        return self
+    def save_json(self, path=None):
+        path = Save.json(obj=self.__dict__, path=path)
+        return path
 
-    def __bool__(self):
-        return bool(self.__dict__)
+    # def save_yaml(self, path=None) -> str:
+    #     return Save.yaml(self.__dict__, str(path))
 
     @staticmethod
     def recursive_struct(mydict):
@@ -873,9 +897,32 @@ class Struct(dict):  # inheriting from dict gives `get` method.
             default_ = [None] * len(names)
         return cls.from_keys_values(names, values=default_)
 
-    def get_values(self, keys):
-        return List([self[key] for key in keys])
+    def spawn_from_values(self, values):
+        """From the same keys, generate a new Struct with different values passed."""
+        return self.from_keys_values(self.keys(), self.evalstr(values, expected='self'))
 
+    def spawn_from_keys(self, keys):
+        """From the same values, generate a new Struct with different keys passed."""
+        return self.from_keys_values(self.evalstr(keys, expected="self"), self.values())
+
+    def to_default(self, default=lambda: None):
+        from collections import defaultdict
+        tmp2 = defaultdict(default)
+        tmp2.update(self.__dict__)
+        self.__dict__ = tmp2
+        return self
+
+    # ================ magic ==========================
+    def __bool__(self):
+        return bool(self.__dict__)
+
+    def __contains__(self, key):
+        return key in self.__dict__
+
+    def __len__(self):
+        return len(self.keys())
+
+    # =========================== print ===========================
     @property
     def clean_view(self):
 
@@ -888,13 +935,18 @@ class Struct(dict):  # inheriting from dict gives `get` method.
 
     def __repr__(self):
         repr_string = ""
-        for key in self.keys().list:
+        for key in self.keys().to_list():
             repr_string += str(key) + ", "
         return "Struct: [" + repr_string + "]"
 
-    def print(self, sep=None, yaml=False, dtype=True, logger=False, limit=50, config=False, newline=True):
+    def print(self, sep=None, yaml=False, dtype=True, return_str=False, limit=50, config=False, newline=True):
         if config:
-            return print(DisplayData.config(self.__dict__, newline=newline))
+            repr_str = Display.config(self.__dict__, newline=newline)
+            if return_str: return repr_str
+            else:
+                print(repr_str)
+                return self
+
         if bool(self) is False:
             print(f"Empty Struct.")
             return None  # break out of the function.
@@ -913,12 +965,12 @@ class Struct(dict):  # inheriting from dict gives `get` method.
         for key in self.keys().list:
             key_str = str(key)
             type_str = str(type(self[key])).split("'")[1]
-            val_str = DisplayData.get_repr(self[key], limit=limit).replace("\n", " ")
+            val_str = Display.get_repr(self[key], limit=limit).replace("\n", " ")
             repr_string += key_str + " " * abs(sep - len(key_str)) + " " * len("Key")
             if dtype:
                 repr_string += type_str + " " * abs(sep - len(type_str)) + " " * len("Item Type")
             repr_string += val_str + "\n"
-        if logger:
+        if return_str:
             return repr_string
         else:
             print(repr_string)
@@ -959,18 +1011,6 @@ class Struct(dict):  # inheriting from dict gives `get` method.
     def __iter__(self):  # used when list(~) is called or it is iterated over.
         return iter(self.dict.items())
 
-    @staticmethod
-    def save_yaml(path):
-        Save.yaml(path)
-
-    @property
-    def dict(self):  # allows getting dictionary version without accessing private memebers explicitly.
-        return self.__dict__
-
-    @dict.setter
-    def dict(self, adict):
-        self.__dict__ = adict
-
     def update(self, *args, **kwargs):
         """Accepts dicts and keyworded args
         """
@@ -1023,30 +1063,33 @@ class Struct(dict):  # inheriting from dict gives `get` method.
                             total_dict[key] = adict[key]
         return Struct(total_dict)
 
-    def keys(self):
+    def keys(self) -> List:
         """Same behaviour as that of `dict`, except that is doesn't produce a generator."""
-        return List(self.dict.keys())
+        return List(list(self.dict.keys()))
 
-    def values(self):
+    def values(self) -> List:
         """Same behaviour as that of `dict`, except that is doesn't produce a generator."""
-        return List(self.dict.values())
+        return List(list(self.dict.values()))
 
-    def items(self):
+    def items(self) -> List:
         """Same behaviour as that of `dict`, except that is doesn't produce a generator."""
         return List(self.dict.items())
 
+    def get_values(self, keys) -> List:
+        return List([self[key] for key in keys])
+
     def to_dataframe(self, *args, **kwargs):
         # return self.values().to_dataframe(names=self.keys())
         import pandas as pd
         return pd.DataFrame(self.__dict__, *args, **kwargs)
 
-    def spawn_from_values(self, values):
-        """From the same keys, generate a new Struct with different values passed."""
-        return self.from_keys_values(self.keys(), self.evalstr(values, expected='self'))
+    @property
+    def dict(self):  # allows getting dictionary version without accessing private memebers explicitly.
+        return self.__dict__
 
-    def spawn_from_keys(self, keys):
-        """From the same values, generate a new Struct with different keys passed."""
-        return self.from_keys_values(self.evalstr(keys, expected="self"), self.values())
+    @dict.setter
+    def dict(self, adict):
+        self.__dict__ = adict
 
     def plot(self, artist=None):
         if artist is None:
@@ -1065,11 +1108,11 @@ class Struct(dict):  # inheriting from dict gives `get` method.
         return artist
 
 
-class DisplayData:
+class Display:
     # primitive = {list, dict, tuple}
 
     @staticmethod
-    def set_pandas_display(rows=1000, columns=1000, width=1000, colwidth=40):
+    def set_pandas_display(rows=1000, columns=1000, width=5000, colwidth=40):
         import pandas as pd
         pd.set_option('display.max_colwidth', colwidth)
         pd.set_option('display.max_columns', columns)  # to avoid replacing them with ...
@@ -1115,7 +1158,7 @@ class DisplayData:
     def config(mydict, newline=True):
         rep = ""  # returns json repr
         for key, val in mydict.items():
-            rep += f"{key}={val}"
+            rep += f"{key} = {val}"
             rep += "\n" if newline else ", "
         return rep
 
@@ -1136,4 +1179,5 @@ class DisplayData:
 
 
 if __name__ == '__main__':
+    # a = 2
     pass
