@@ -25,11 +25,11 @@ class HyperParam(tb.Struct):
     """
     subpath = tb.P('metadata/hyper_params')  # location within model directory where this will be saved.
 
-    def __init__(self, *args, **kwargs):
+    def __init__(self, **kwargs):
         super().__init__(
             # ==================== Enviroment =========================
-            exp_name='default',
-            root=tb.tmp(folder="tmp_models"),
+            name='default_model_name_' + tb.randstr(),
+            root=tb.P.tmp(folder="tmp_models"),
             pkg_name='tensorflow',
             device_name=Device.gpu0,
             # ===================== Data ==============================
@@ -37,32 +37,37 @@ class HyperParam(tb.Struct):
             shuffle=True,
             precision='float32',
             # ===================== Model =============================
+            # depth = 3
             # ===================== Training ==========================
-            split=0.2,  # test split
-            lr=0.0005,
+            test_split=0.2,  # test split
+            learning_rate=0.0005,
             batch_size=32,
             epochs=30,
         )
         self._configured = False
         self.device_name = None
-        # self.update(*args, **kwargs)
-        # self.save_code()
-        # self.config_device()
+        self.save_type = ["data", "whole", "both"][-1]
+        self.update(**kwargs)
 
     @property
     def save_dir(self):
-        return self.root / self.exp_name
+        return tb.P(self.root) / self.name
 
-    def save(self):
-        self.save_dir.joinpath(self.subpath / '.txt').write_text(data=str(self))
-        super(HyperParam, self).save(path=self.save_dir.joinpath(self.subpath / ".pkl"))
+    def save(self, path=None, itself=True, r=False, include_code=False, add_suffix=True):
+        self.save_dir.joinpath(self.subpath / 'hparams.txt').create(parent_only=True).write_text(data=str(self))
+        if self.save_type in {"whole", "both"}:
+            super(HyperParam, self).save(path=self.save_dir.joinpath(self.subpath / "hparams.HyperParam.pkl"),
+                                         itself=True, add_suffix=False)
+        if self.save_type in {"data", "both"}:
+            super(HyperParam, self).save(path=self.save_dir.joinpath(self.subpath) / "hparams.HyperParam.dat.pkl",
+                                         itself=False, add_suffix=False)
 
     @classmethod
-    def from_saved(cls, path):
-        return (tb.P(path) / HyperParam.subpath / ".pkl").readit()
+    def from_saved(cls, path, *args, r=False, scope=None, **kwargs):
+        return super(HyperParam, cls).from_saved(path=tb.P(path) / cls.subpath / "hparams.HyperParam.dat.pkl")
 
     def __repr__(self):
-        self.print(config=True)
+        return tb.Struct(self.__dict__).print(config=True, return_str=True)
 
     @property
     def pkg(self):
@@ -163,7 +168,7 @@ class HyperParam(tb.Struct):
             print(f"Device already configured, skipping ... ")
 
 
-class DataReader(object):
+class DataReader(tb.Base):
     subpath = tb.P("metadata/data_reader")
     """This class holds the dataset for training and testing. However, it also holds meta data for preprocessing
     and postprocessing. The latter is essential at inference time, but the former need not to be saved. As such,
@@ -175,14 +180,37 @@ class DataReader(object):
     def __init__(self, hp: HyperParam = None, specs=None, split=None, *args, **kwargs):
         super().__init__(*args, **kwargs)
         self.hp = hp
-        self.specs = specs if specs else tb.Struct()  # Summary of data to be memorized by model
         self.split = split
         self.plotter = None
+        # attributes to be saved.
+        self.specs = specs if specs else tb.Struct()
+        self.scaler = None
+
+    def save(self, path=None, *args, **kwargs):
+        base = (tb.P(path) if path is not None else self.hp.save_dir).joinpath(self.subpath)
+        if self.hp.save_type in {"whole", "both"}:
+            super(DataReader, self).save(path=base / "data_reader.DataReader.pkl", itself=True, add_suffix=False)
+        if self.hp.save_type in {"data", "both"}:
+            super(DataReader, self).save(path=base / "data_reader.DataReader.dat.pkl", itself=False, add_suffix=False)
+
+    @classmethod
+    def from_saved(cls, path, *args, **kwargs):
+        instance = cls(*args, **kwargs)
+        data = (tb.P(path) / cls.subpath / "data_reader.DataReader.dat.pkl").readit()
+        instance.__setstate__(data)
+        return instance
+
+    def __getstate__(self):
+        return dict(specs=self.specs, scaler=self.scaler)
+
+    def __setstate__(self, state):
+        """hp is miassing, deliberate by design."""
+        return self.__dict__.update(state)
 
     def __repr__(self):
-        return f"DataReader Object with these keys: \n" + self.specs.keys().__str__()
+        return f"DataReader Object with these keys: \n" + tb.Struct(self.__dict__).print(config=True, return_str=True)
 
-    def data_split(self, *args, strings=None, **kwargs):
+    def split_the_data(self, *args, strings=None, **kwargs):
         """
         :param args: whatever to be sent to train_test_split
         :param kwargs: whatever to be sent to train_test_split
@@ -191,7 +219,7 @@ class DataReader(object):
         """
         # import sklearn.preprocessing as preprocessing
         from sklearn.model_selection import train_test_split
-        result = train_test_split(*args, test_size=self.hp.split, shuffle=self.hp.shuffle,
+        result = train_test_split(*args, test_size=self.hp.test_split, shuffle=self.hp.shuffle,
                                   random_state=self.hp.seed, **kwargs)
         self.split = tb.Struct(train_loader=None, test_loader=None)
         if strings is None:
@@ -223,15 +251,6 @@ class DataReader(object):
         op = np.random.randn(*((self.hp.batch_size,) + op_shape)).astype(dtype)
         return ip, op
 
-    def save(self):
-        tb.Save.pickle(self.specs, path=self.hp.save_dir.joinpath(self.subpath / "data_reader.pkl"))
-
-    @classmethod
-    def from_saved(cls, path, *args, **kwargs):
-        instance = cls(*args, **kwargs)
-        instance.specs = (tb.P(path) / cls.subpath / "data_reader.pkl").readit()
-        return instance
-
     def preprocess(self, *args, **kwargs):
         _ = args, kwargs, self
         return args[0]  # acts like identity.
@@ -240,6 +259,13 @@ class DataReader(object):
         _ = args, kwargs, self
         return args[0]  # acts like identity
 
+    def standardize(self):
+        assert self.split is not None, "Load up the data first."
+        from sklearn.preprocessing import StandardScaler
+        self.scaler = StandardScaler()
+        self.split.x_train = self.scaler.fit_transform(self.split.x_train)
+        self.split.x_test = self.scaler.transform(self.split.x_test)
+
     def image_viz(self, pred, gt=None, names=None, **kwargs):
         """
         Assumes numpy inputs
@@ -299,14 +325,14 @@ class BaseModel(ABC):
             if loss is None:
                 loss = pkg.keras.losses.MeanSquaredError()
             if optimizer is None:
-                optimizer = pkg.keras.optimizers.Adam(self.hp.lr)
+                optimizer = pkg.keras.optimizers.Adam(self.hp.learning_rate)
             if metrics is None:
                 metrics = tb.List()  # [pkg.keras.metrics.MeanSquaredError()]
         elif self.hp.pkg_name == 'torch':
             if loss is None:
                 loss = pkg.nn.MSELoss()
             if optimizer is None:
-                optimizer = pkg.optim.Adam(self.model.parameters(), lr=self.hp.lr)
+                optimizer = pkg.optim.Adam(self.model.parameters(), lr=self.hp.learning_rate)
             if metrics is None:
                 metrics = tb.List()  # [tmp.MeanSquareError()]
         # Create a new compiler object
@@ -338,9 +364,9 @@ class BaseModel(ABC):
         #     self.model.reset_metrics()
         print(f'Switching the optimizer to SGD. Loss is fixed to {self.compiler.loss}'.center(100, '*'))
         if self.hp.pkg.__name__ == 'tensorflow':
-            new_optimizer = self.hp.pkg.keras.optimizers.SGD(lr=self.hp.lr * 0.5)
+            new_optimizer = self.hp.pkg.keras.optimizers.SGD(lr=self.hp.learning_rate * 0.5)
         else:
-            new_optimizer = self.hp.pkg.optim.SGD(self.model.parameters(), lr=self.hp.lr * 0.5)
+            new_optimizer = self.hp.pkg.optim.SGD(self.model.parameters(), lr=self.hp.learning_rate * 0.5)
         self.compiler.optimizer = new_optimizer
         return self.fit(epochs=epochs)
 
@@ -459,7 +485,7 @@ class BaseModel(ABC):
         return tb.pd.DataFrame(loss_dict)
 
     def save_model(self, directory):
-        self.model.save(directory)  # In TF: send only path path. Save path is saved_model.pb
+        self.model.save(directory)  # In TF: send only path dir. Save path is saved_model.pb
 
     def save_weights(self, directory):
         self.model.save_weights(directory.joinpath(self.model.name))  # TF: last part of path is file path.
@@ -497,14 +523,21 @@ class BaseModel(ABC):
         print(f'Model class saved successfully!, check out: {self.hp.save_dir.as_uri()}')
 
     @classmethod
-    def from_class_weights(cls, path, hparam_class, data_class, device_name=None):
+    def from_class_weights(cls, path, hparam_class=None, data_class=None, device_name=None):
         path = tb.P(path)
-        hp_obj = hparam_class.from_saved(path)
+
+        if hparam_class is not None: hp_obj = hparam_class.from_saved(path)
+        else: hp_obj = (path / HyperParam.subpath + ".HyperParam.pkl").readit()
         if device_name: hp_obj.device_name = device_name
-        d_obj = data_class.from_saved(path, hp=hp_obj)
+
+        if data_class is not None: d_obj = data_class.from_saved(path, hp=hp_obj)
+        else: d_obj = (path / DataReader.subpath / "data_reader.DataReader.pkl").readit()
+        d_obj.hp = hp_obj
+
         model_obj = cls(hp_obj, d_obj)
         model_obj.load_weights(path.search('*_save_*')[0])
         model_obj.history = (path / "metadata/history.pkl").readit(notfound=tb.L())
+
         print(f"Class {model_obj.__class__} Loaded Successfully.")
         return model_obj
 
@@ -585,7 +618,7 @@ class Ensemble(tb.Base):
             print("Creating Models".center(100, "="))
             for i in tqdm(range(size)):
                 hp = self.hp_class()
-                hp.exp_name = str(hp.exp_name) + f'__model__{i}'
+                hp.name = str(hp.name) + f'__model__{i}'
                 datacopy = tb.copy.copy(self.data)  # shallow copy
                 datacopy.hp = hp
                 self.models.append(model_class(hp, datacopy))
@@ -727,7 +760,7 @@ class KerasOptimizer:
         pass
 
     def tune(self):
-        kt = tb.core.assert_package_installed("kerastuner")
+        kt = tb.core.install_n_import("kerastuner")
         self.tuner = kt.Hyperband(self,
                                   objective='loss',
                                   max_epochs=10,
