@@ -2,7 +2,8 @@
 import logging
 import subprocess
 import time
-from crocodile.core import np, os, sys, timestamp, randstr, str2timedelta, datetime, Save, assert_package_installed
+from crocodile.core import np, os, sys, timestamp, randstr, str2timedelta, datetime, Save,\
+    dill, install_n_import
 from crocodile.file_management import P
 
 
@@ -83,7 +84,7 @@ class Experimental:
 
     @staticmethod
     def profile_memory(command):
-        psutil = assert_package_installed("psutil")
+        psutil = install_n_import("psutil")
         before = psutil.virtual_memory()
         exec(command)
         after = psutil.virtual_memory()
@@ -232,7 +233,7 @@ class Experimental:
             code_string = args_kwargs + code_string
 
         if copy2clipboard:
-            clipboard = assert_package_installed("clipboard")
+            clipboard = install_n_import("clipboard")
             clipboard.copy(code_string)
         if verbose: print(f"code to be run extracted from {func.__name__} \n", code_string, "=" * 100)
         return code_string  # ready to be run with exec()
@@ -273,7 +274,7 @@ class Experimental:
             res += f"{ak.varkw} = " + "{}\n"
 
         if copy2clipboard:
-            clipboard = assert_package_installed("clipboard")
+            clipboard = install_n_import("clipboard")
             clipboard.copy(res)
         if verbose: print("Finished. Paste code now.")
         return res
@@ -321,7 +322,7 @@ class Experimental:
         else:
             raise KeyError(f"The pointer `{pointer}` was not found in the module `{module}`")
         print(cell)
-        clipboard = assert_package_installed("clipboard")
+        clipboard = install_n_import("clipboard")
         clipboard.copy(cell)
         return cell
 
@@ -629,34 +630,36 @@ class Terminal:
             extra = ""
         my_list = []
         if self.machine == "Windows":
-            my_list += [new_window, terminal, shell, extra]
+            my_list += [new_window, terminal, shell, extra]  # by having a list,
+            # it is equivalent to: start "ipython -i file.py". Thus, arguments of ipython go to ipython, not start.
         my_list += cmds
-        print("Meta.Terminal.run_async: Subprocess command: ", my_list)
         my_list = [item for item in my_list if item != ""]
+        print("Meta.Terminal.run_async: Subprocess command: ", my_list)
         w = subprocess.Popen(my_list, stdin=subprocess.PIPE, shell=True)  # stdout=self.stdout, stderr=self.stderr, stdin=self.stdin
         # returns Popen object, not so useful for communcation with an opened terminal
         return w
 
     @staticmethod
     def run_script(script, wdir=None, interactive=True, ipython=True,
-                   shell=None, delete=False, terminal="", new_window=True):
+                   shell=None, delete=False, terminal="", new_window=True, header=True):
         """This method is a wrapper on top of `run_async" except that the command passed will launch python
         terminal that will run script passed by user.
         * Regular Python is much lighter than IPython. Consider using it while not debugging.
         """
+        # TODO: add option whether to add prepend to the script or not.
         wdir = wdir or P.cwd()
-        header = f"""
-# The following lines of code form a header appended by Terminal.run_script
+        header_script = f"""
+# ======================== Code prepended by Terminal.run_script =========================
 import crocodile.toolbox as tb
 tb.sys.path.insert(0, r'{wdir}')
-# End of header, start of script passed:
+# ======================== End of header, start of script passed: ========================
 """  # this header is necessary so import statements in the script passed are identified relevant to wdir.
 
-        script = header + script
+        script = header_script + script if header else script
         if terminal in {"wt", "powershell", "pwsh"}:
             script += "\ntb.DisplayData.set_pandas_auto_width()\n"
         script = f"""print(r'''{script}''')""" + "\n" + script
-        file = P.tmpfile(name="tmp_python_script", suffix=".py", folder="tmpscripts")
+        file = P.tmpfile(name="tmp_python_script", suffix=".py", folder="tmp_scripts")
         file.write_text(script)
         print(f"Script to be executed asyncronously: ", file.absolute().as_uri())
         Terminal().run_async(f"{'ipython' if ipython else 'python'}",
@@ -669,30 +672,44 @@ tb.sys.path.insert(0, r'{wdir}')
         # command = f'ipython {"-i" if interactive else ""} -c "{script}"'
 
     @staticmethod
-    def load_object_in_new_session(obj):
+    def replicate_in_new_session(obj, execute=False):
         """Python brachnes off to a new window and run the function passed.
         context can be either a pickled session or the current file __file__"""
         # step 1: pickle the function
         # step 2: create a script that unpickles it.
         # step 3: run the script that runs the function.
-        # TODO complete this
-        fname = P.tmpfile(tstamp=False, suffix=".pkl")
-        Save.pickle(obj=obj, path=fname, verbose=False)
+        file = P.tmpfile(tstamp=False, suffix=".pkl")
+        Save.pickle(obj=obj, path=file, verbose=False)
         script = f"""
-path = tb.P(r'{fname}')
+path = tb.P(r'{file}')
 obj = path.readit()
 path.delete(sure=True, verbose=False)
+obj{'()' if execute else ''}
 """
         Terminal.run_script(script)
 
+    @staticmethod
+    def replicate_session(cmd=""):
+        file = P.tmpfile(suffix=".pkl")
+        dill.dump_session(file, main=sys.modules[__name__])
+        script = f"""
+path = tb.P(r'{file}')
+tb.dill.load_session(str(path)); 
+path.delete(sure=True, verbose=False)
+{cmd}
+"""
+        Terminal().run_script(script=script)
+
 
 class SSH(object):
-    def __init__(self, hostname, username, ssh_key=None):
+    def __init__(self, username, hostname, ssh_key=None, pwd=None):
         _ = False
         if _:
             super().__init__()
+        self.ssh_key = str(ssh_key) if ssh_key is not None else None
+        # no need to pass ssh_key if it was configured properly already
+
         import paramiko
-        self.ssh_key = ssh_key
         self.ssh = paramiko.SSHClient()
         self.ssh.load_system_host_keys()
         self.ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
@@ -700,13 +717,23 @@ class SSH(object):
         self.username = username
         self.ssh.connect(hostname=hostname,
                          username=username,
-                         port=22, key_filename=self.ssh_key.string if self.ssh_key is not None else None)
+                         password=pwd,
+                         port=22, key_filename=self.ssh_key)
         self.sftp = self.ssh.open_sftp()
-        self.load_python_cmd = rf"""source ~/miniconda3/bin/activate"""  # possible activate an env
+        self.target_machine = "Windows" if self.run("$env:OS", printit=False).output["stdout"] == "Windows_NT" else "Linux"
+        # it must uses a python independent way to figure out the machine type to avoid circualrity below:
         import platform
         self.platform = platform
-        self.target_machine = self.ssh.exec_command(self.load_python_cmd +
-                                                    "python -c 'import platform; platform.system()'")
+
+        if self.platform.system() == "Windows":
+            self.local_python_cmd = rf"""~/venvs/ve/Scripts/activate"""  # works for both cmd and pwsh
+        else:
+            self.local_python_cmd = rf"""source ~/venvs/ve/bin/activate"""
+
+        if self.target_machine == "Windows":
+            self.remote_python_cmd = rf"""~/venvs/ve/Scripts/activate"""  # works for both cmd and pwsh
+        else:
+            self.remote_python_cmd = rf"""source ~/venvs/ve/bin/activate"""
 
     def get_key(self):
         """In SSH commands you need this:
@@ -717,8 +744,14 @@ class SSH(object):
         else:
             return ""
 
+    @staticmethod
+    def copy_ssh_keys_to_remote(fqdn):
+        # Caveat, only windows is supported.
+        # the following is a Windows Openssh alternative to ssh-copy-id
+        Terminal().run(f'type $env:USERPROFILE\.ssh\id_rsa.pub | ssh {fqdn} "cat >> .ssh/authorized_keys"')
+
     def __repr__(self):
-        return f"{self.local()} SSH connection to {self.remote()}"
+        return f"{self.local()} [{self.platform.system()}] SSH connection to {self.remote()} [{self.target_machine}] "
 
     def remote(self):
         return f"{self.username}@{self.hostname}"
@@ -734,33 +767,39 @@ class SSH(object):
         assert self.target_machine == "Linux"
         self.run(f"{name} = {os.environ[name]}; export {name}")
 
-    def copy_from_here(self, source, target=None, zip_n_encrypt=True):
+    def copy_from_here(self, source, target=None, zip_n_encrypt=False):
         pwd = None
         if zip_n_encrypt:
             pwd = randstr(length=10, safe=True)
             source = P(source).expanduser().zip_n_encrypt(pwd=pwd)
+
+        if target is None:
+            target = P(source).collapseuser()
+            assert target.is_relative_to("~"), f"If target is not specified, source must be relative to home."
+            target = target.as_posix()
         print(f"Step 1: Creating Target directory {target} @ remote machine.")
-        if target is None: target = P(source).collapseuser().as_posix()  # works if source is relative.
+
         resp = self.runpy(f'print(tb.P(r"{target}").expanduser().parent.create())')
         remotepath = P(resp.op or "").joinpath(P(target).name).as_posix()
-        print(f"Send `{source}` ==> `{remotepath}` ")
+        print(f"SENT `{source}` ==> `{remotepath}`")
         self.sftp.put(localpath=P(source).expanduser(), remotepath=remotepath)
         if zip_n_encrypt:
             resp = self.runpy(f"""tb.P(r"{remotepath}").expanduser().decrypt_n_unzip(pwd="{pwd}", inplace=True)""")
+            source.delete(sure=True)
             return resp
 
     def copy_to_here(self, source, target=None):
         pass
 
     def run(self, cmd, printit=True):
-        print(f"\nExecuting on remote {self.username}@{self.hostname}:\n{cmd}")
+        # if printit: print(f"\nExecuting on remote {self.username}@{self.hostname}:\n{cmd}")
         res = self.ssh.exec_command(cmd)
         res = Terminal.Response(stdin=res[0], stdout=res[1], stderr=res[2], cmd=cmd)
         if printit: res.print()
         return res
 
     def runpy(self, cmd):
-        cmd = f"""{self.load_python_cmd}; python -c 'import crocodile.toolbox as tb; {cmd} ' """
+        cmd = f"""{self.remote_python_cmd}; python -c 'import crocodile.toolbox as tb; {cmd} ' """
         return self.run(cmd)
 
     def run_locally(self, command):
@@ -774,13 +813,13 @@ class Log(object):
      is printing this message, in addition to many other concerns.
 
      Advantages of using instances of this class: You do not need to worry about object pickling process by modifing
-     the __getstate__ method of the class that will own the logger. This is the case because loggers lose access
-     to the file logger when unpickled, so it is better to instantiate them again.
+     the __getstate__ method of the class that will own the return_str. This is the case because loggers lose access
+     to the file return_str when unpickled, so it is better to instantiate them again.
      Logger can be pickled, but its handlers are lost, so what's the point? no perfect reconstruction.
      Additionally, this class keeps track of log files used, append to them if they still exist.
 
      Implementation detail: the design favours composition over inheritence. To counter the inconvenience
-      of having extra typing to reach the logger, a property `logger` was added to Base class to refer to it."""
+      of having extra typing to reach the return_str, a property `return_str` was added to Base class to refer to it."""
 
     def __init__(self, dialect=["colorlog", "logging", "coloredlogs"][0],
                  name=None, file: bool = False, file_path=None, stream=True, fmt=None, sep=" | ",
@@ -793,15 +832,11 @@ class Log(object):
         self.verbose = verbose  # specific to coloredlogs dialect
         self.log_colors = log_colors  # specific kwarg to colorlog dialect
         self.owners = []  # list of objects using this object to log. It won't be pickled anyway, no circularity prob
-        if file is False and stream is False:
-            self.logger = Null()
-        else:
-            self.logger = Null()  # to be populated by `_install`
-            self._install()
-            # update specs after intallation.
-            self.specs["path"] = self.logger.name
-            if file:  # first handler is a file handler
-                self.specs["file_path"] = self.logger.handlers[0].baseFilename
+        self._install()
+        # update specs after intallation.
+        self.specs["path"] = self.logger.name
+        if file:  # first handler is a file handler
+            self.specs["file_path"] = self.logger.handlers[0].baseFilename
 
     def __getattr__(self, item):  # makes it twice as slower as direct access 300 ns vs 600 ns
         return getattr(self.logger, item)
@@ -821,8 +856,8 @@ class Log(object):
     def critical(self, msg):
         return self.logger.critical(msg)
 
-    def set_level(self, level, which=["logger", "stream", "file", "all"][0]):
-        if which in {"logger", "all"}: self.logger.setLevel(level)
+    def set_level(self, level, which=["return_str", "stream", "file", "all"][0]):
+        if which in {"return_str", "all"}: self.logger.setLevel(level)
         if which in {"stream", "all"}: self.get_shandler().setLevel(level)
         if which in {"file", "all"}: self.get_fhandler().setLevel(level)
 
@@ -846,8 +881,10 @@ class Log(object):
     def file(self):
         return P(self.specs["file_path"]) if self.specs["file_path"] else None
 
-    def _install(self):  # populates self.logger attribute according to specs and dielect.
-        if self.dialect == "colorlog":
+    def _install(self):  # populates self.return_str attribute according to specs and dielect.
+        if self.specs["file"] is False and self.specs["stream"] is False:
+            self.logger = Null()
+        elif self.dialect == "colorlog":
             self.logger = Log.get_colorlog(log_colors=self.log_colors, **self.specs)
         elif self.dialect == "logging":
             self.logger = Log.get_logger(**self.specs)
@@ -858,20 +895,24 @@ class Log(object):
 
     def __setstate__(self, state):
         self.__dict__ = state
-        if self.specs["file_path"] is not None:
-            self.specs["file_path"] = P.home() / self.specs["file_path"]
+        if self.specs["file_path"] is not None:  # this way of creating relative path makes transferrable
+            # across machines.
+            self.specs["file_path"] = P(self.specs["file_path"]).rel2home()
         self._install()
 
     def __getstate__(self):
-        # logger can be pickled without this method,
+        # return_str can be pickled without this method,
         # but its handlers are lost, so what's the point? no perfect reconstruction.
         state = self.__dict__.copy()
         state["specs"] = state["specs"].copy()
-        del state["logger"]
+        del state["return_str"]
         if self.specs["file_path"] is not None:
-            state["specs"]["file_path"] = P(self.specs["file_path"]).rel2home()
+            state["specs"]["file_path"] = P(self.specs["file_path"]).expanduser()
         return state
 
+    def close(self):  # close file.
+        raise NotImplementedError
+
     def __repr__(self):
         tmp = f"{self.logger} with handlers: \n"
         for h in self.logger.handlers:
@@ -909,9 +950,9 @@ class Log(object):
                         'path': {'color': 'blue'},
                         'programname': {'color': 'cyan'},
                         'username': {'color': 'yellow'}}
-        coloredlogs = assert_package_installed("coloredlogs")
+        coloredlogs = install_n_import("coloredlogs")
         if verbose:
-            verboselogs = assert_package_installed("verboselogs")
+            verboselogs = install_n_import("verboselogs")
             # https://github.com/xolox/python-verboselogs
             # verboselogs.install()  # hooks into logging module.
             logger = verboselogs.VerboseLogger(name=name)
@@ -938,7 +979,7 @@ class Log(object):
                           'ERROR': 'thin_red',
                           'CRITICAL': 'fg_bold_red,bg_white',
                           }  # see here for format: https://pypi.org/project/colorlog/
-        colorlog = assert_package_installed("colorlog")
+        colorlog = install_n_import("colorlog")
         logger = Log.get_base_logger(colorlog, name, l_level)
         fmt = colorlog.ColoredFormatter(fmt or (rf"%(log_color)s" + Log.get_format(sep)), log_colors=log_colors)
         Log.add_handlers(logger, colorlog, file, f_level, file_path, fmt, stream, s_level)
@@ -947,7 +988,7 @@ class Log(object):
     @staticmethod
     def get_logger(name=None, file=False, file_path=None, stream=True, fmt=None, sep=" | ",
                    s_level=logging.DEBUG, f_level=logging.DEBUG, l_level=logging.DEBUG):
-        """Basic Python logger."""
+        """Basic Python return_str."""
         logger = Log.get_base_logger(logging, name, l_level)
         fmt = logging.Formatter(fmt or Log.get_format(sep))
         Log.add_handlers(logger, logging, file, f_level, file_path, fmt, stream, s_level)
@@ -965,9 +1006,9 @@ class Log(object):
 
     @staticmethod
     def add_handlers(logger, module, file, f_level, file_path, fmt, stream, s_level):
-        if file or file_path:  # ==> create file handler for the logger.
+        if file or file_path:  # ==> create file handler for the return_str.
             Log.add_filehandler(logger, file_path=file_path, fmt=fmt, f_level=f_level)
-        if stream:  # ==> create stream handler for the logger.
+        if stream:  # ==> create stream handler for the return_str.
             Log.add_streamhandler(logger, s_level, fmt, module=module)
 
     @staticmethod
@@ -982,7 +1023,7 @@ class Log(object):
     @staticmethod
     def add_filehandler(logger, file_path=None, fmt=None, f_level=logging.DEBUG, mode="a", name="myFileHandler"):
         if file_path is None:
-            file_path = P.tmpfile(name="logger", suffix=".log", folder="loggers")
+            file_path = P.tmpfile(name="return_str", suffix=".log", folder="loggers")
         fhandler = logging.FileHandler(filename=str(file_path), mode=mode)
         fhandler.setFormatter(fmt=fmt)
         fhandler.setLevel(level=f_level)
@@ -1144,7 +1185,7 @@ class Scheduler:
 
 
 def qr(txt):
-    qrcode = assert_package_installed("qrcode")
+    qrcode = install_n_import("qrcode")
     img = qrcode.make(txt)
     file = P.tmpfile(suffix=".png")
     img.save(file.__str__())
